#!/usr/bin/env python
# -*- coding: utf-8 -*-

# python import
import os
from functools import partial

# json tool
import simplejson as json

# twisted import
from twisted.python import log

# nevow import
from nevow import inevow, loaders, rend, static, tags as T
# nevow i18n import
from nevow import i18n
from nevow.i18n import _

# bewype ui import
from bewype.core import ui, utils
# bewpe cache import
from bewype.core.cache import PageManager
# bewpe config import
from bewype.core.config import Config
# bewpe component import
from bewype.core.component import Layout
# bewpe db import
from bewype.core.db import Query
# bewype message import
from bewype.core.message import RpcResult
# bewpe resource import
from bewype.core.resource import Path, StaticUpdater
# bewype js import
import bewype.js

DOCTYPE = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0
    Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
"""

class Page(rend.Page):
    """Page object for the bewype application. Manage headers, current language
    code, layout and children widgets to render and `nevow` `locateChild` method
    for ajax calls or html rendering dispatching.
    """

    addSlash = True

    docFactory = loaders.stan(
            [
                T.raw(DOCTYPE),
                T.html[
                    T.head(render=T.directive("headers")),
                    T.body(render=T.directive("widgets"))
                    ]
                ]
            )

    def __init__(self, page_url=None, realm=None, language_code=None,
            init_headers=True):
        """Creates a new page object for the passed url. Children layouts and
        widgets are initialized according to the corresponding `Page` db item.

        If realm is set by a `nevow` secure portal, the object will be able to
        return the avatar id of authenticated user.

        :param page_url: url of the page, default /index
        :param realm: realm object from the `nevow` portal, None if in public mode
        :param language_code: language code of the page
        :param init_headers: can be set to False for unit testing.
        """
        # init parent
        rend.Page.__init__(self)
        # set realm
        self.realm = realm
        # init locale settings
        self.set_language(language_code)
        # ensure default index page
        if page_url is None:
            self.page_url = "/index"
        else:
            self.page_url = page_url
        # init lists
        self.widgets = list()
        # init dicts
        self.children = dict()
        self.children_callbacks = dict()
        # init header list
        self.headers = list()
        # init ...
        self._init_locale_settings()
        # do not init header for test but only for served pages
        if init_headers is True:
            self._init_headers()
        else:
            pass
        self._init_layout()

    def set_language(self, language_code):
        """Set/updates the page language code or use the default config
        value if `language_code` value is None.

        :param language_code: new page language code ('en', 'fr', Etc.) or None
        """
        # set/ensure class var
        if language_code is None:
            self.language_code = Config().get("app>language")
        else:
            self.language_code = language_code

    def _init_locale_settings(self):
        # get resource path
        _resource_path = Config().resource_path()
        # little check
        if _resource_path is None:
            pass
        # prepare i18n config
        else:
            _locale_dir = os.path.join(_resource_path, "i18n")
            _i18nConfig = i18n.I18NConfig(domain="bewype", localeDir=_locale_dir)
            # do you remember
            self.remember(_i18nConfig, inevow.II18NConfig)

    def _init_layout(self):
        # get page from db
        _page_item = Query("Page").get_item("page_url", self.page_url)
        if _page_item is None:
            return
        else:
            # build and get the main widget of the page
            _layout_component = Layout(self, _page_item.layout)
            # add the main widget to the page
            self._add_widget(_layout_component)

    def _append_page_title(self):
        # get page db item
        _page = Query("Page").get_item("page_url", self.page_url)
        # little check
        if _page is None:
            pass
        else:
            # get page title
            _page_title = _page.get_page_title(self.language_code)
            # second check
            if _page_title is None:
                pass
            else:
                # set page title
                self._add_header(T.title[_page_title])

    def _append_js_lib(self, js_lib, js_file_name):
        # prepare path to add
        _res_path = os.path.join("script", js_lib, "%s.js" % js_file_name)
        # get updater
        _s_updater = StaticUpdater(self, py_file=bewype.js.__file__)
        # do update
        _s_updater.append_static_resource(_res_path, use_type=False)

    def _append_js_dir(self, js_lib):
        # get __modules path
        _m_path = os.path.join("script", js_lib, "__modules")
        # get module file and read it
        _f_path = Path().get_resource_path_from_py_file(bewype.js.__file__,
                sub_path=_m_path)
        # get modules list
        _f = open(_f_path)
        _lines = _f.readlines()
        _f.close()
        # do adds
        map(lambda _fn: self._append_js_lib(js_lib, _fn.strip()), _lines)

    def _append_static_dir(self, dir_path):
        # do dir add
        _u = StaticUpdater(self)
        _u.append_static_dir(dir_path=dir_path)
        # bewype js resources
        _u = StaticUpdater(self, py_file=bewype.js.__file__)
        _u.append_static_dir(dir_path=dir_path)

    def _init_headers(self):
        """
        """
        # add title
        self._append_page_title()
        # prepare js libs to add
        _js_libs = [
                "jQuery",
                "MochiKit",
                "MooTools",
                "MooEditable",
                "Other",
                "BewypeJS",
                "BewypeJS.Utils",
                "BewypeJS.Rpc",
                "BewypeJS.Ui",
                "BewypeJS.Component"
                ]
        map(self._append_js_dir, _js_libs)
        # do map to append static dirs
        map(self._append_static_dir, utils.get_resource_types())
        # deliver or not ??
        if Config().deliver_static_files():
            # do not deliver static behind a production server, ex.: nginx
            _static_path = Config().static_path()
            setattr(self, "child_static", static.File(_static_path))
        else:
            pass

    def is_protected(self):
        """Returns True is use a `Realm` object as in secure guarded mode.
        """
        return self.realm is not None

    def get_avatar_id(self):
        """Returns the avatar id of the current authenticated user using the
        realm object.
        """
        if self.realm is None:
            return None
        else:
            return self.realm.avatarId

    def logout(self):
        """`Not used logout callback.`
        """
        # get avatar id
        _avatar_id = self.get_avatar_id()
        # DEBUG
        # log.msg("[generic_page] logout - _avatar_id: %s" % _avatar_id)

    def _service_call(self, context, page, ajax_key):
        """Dispatched ajax calls using the children callbacks dict of the page.
        Each callback should return a RpcResult to be serialized as json
        response.

        :param context: nevow current context
        :param page: requested page accroding to the requested url
        :param ajax_key: key found in the request
        """
        # get js callback
        _w_element, _service_cb = page.children_callbacks[ajax_key]
        # get current avatar id
        _avatar_id = self.get_avatar_id()
        # get user db item
        _user_item = Query("User").get_item("user_name", _avatar_id)
        # check user widget permission
        if _user_item is None:
            _u_has_permissions = False
        else:
            _u_has_permissions = _user_item.has_widget_permission(_w_element)
        # get public flag
        _public = len(_w_element.permissions) == 0
        # check permission
        if _u_has_permissions \
        or _public:
            # get request args
            _data = inevow.IRequest(context).args
            # show template dict in debug
            # -> don't catch errors to get the full trace
            if Config().get("app>debug", type_=bool) is True:
                # ajax callback
                _rpc_result = _service_cb(_data, public=_public)
            # basic blind catch for prod environment (should not happen :s)
            else:
                try:
                    # ajax callback
                    _rpc_result = _service_cb(_data, public=_public)
                except Exception, e:
                    # error in ajax call back
                    log.err("[gen_page] _service_call - e: %s" % e)
                    # temp message ?? never appear ??
                    _rpc_result = RpcResult()
                    _rpc_result.error(_("Oops!"))
        # not authorized message
        else:
            # log unexpected avatar id
            log.err("[gen_page] _service_call - not authorized: %s" \
                    % _avatar_id)
            # not authorized message
            _rpc_result = RpcResult()
            _rpc_result.error(_("Restricted Acess!"))
        # render result
        return json.dumps(_rpc_result.serialize(context))

    def locateChild(self, context, segments):
        """`Nevow` low level page dispatcher. Parse the given url using passed
        `segments` and call ajax method or render html content according to the
        arguments of the given context.

        :param context: a `nevow` context
        :param segments: `nevow` segments for the requested url
        """
        # call page parent locateChild
        child, remainingSegments = \
                rend.Page.locateChild(self, context, segments)
        # get request object from context
        _request = inevow.IRequest(context)
        # get page url
        _page_url = _request.path
        # get language
        _language_code = segments[0]
        # update url key
        _page_url = "/%s" % "/".join(segments[1:])
        _ajax_url = "/%s" % "/".join(segments[1:-1])
        # get page url list from db        
        _q = Query("Page")
        _url_list = _q.select(column_list=[_q.obj.page_url])
        #
        if _page_url in map(lambda _i: _i[0], _url_list):
            # not an ajax call
            _ajax_key = None
        elif _ajax_url in map(lambda _i: _i[0], _url_list):
            # update page url
            _page_url = _ajax_url
            # get last segment for ajax as ajax cb key
            _ajax_key = segments[-1]
        else:
            return child, remainingSegments
        #
        if _page_url == self.page_url:
            # here is the current page
            _page = self
        elif PageManager().has_page(_page_url):
            # here is the current page
            _page = PageManager().get_page(_page_url)
        else:
            # create the new page
            _page = Page(page_url=_page_url, realm=self.realm)
            # add page to the current dict
            PageManager().add_page(_page_url, _page)
        # update page language
        _page.set_language(_language_code)
        # update context language
        context.remember([_language_code], inevow.ILanguages)
        # found and return page
        if _ajax_key is None or _ajax_key == "":
            return _page, []
        # manage ajax call
        elif _page.children_callbacks.has_key(_ajax_key):
            return self._service_call(context, _page, _ajax_key), []
        # something else
        else:
            return child, remainingSegments

    def renderHTTP(self, context):
        """Override the render HTTP base method to ensure good i18n management.
        If the language code has changed it ensure the page will be flattened
        using the good .mo file.

        :param context: the current `nevow` context
        """
        # update context language
        context.remember([self.language_code], inevow.ILanguages)
        # Let the base class handle it, really.
        return rend.Page.renderHTTP(self, context)

    def _add_header(self, header):
        self.headers.append(header)

    def render_headers(self, context, data):
        """Returns the list of the current prepared page headers.
        """
        return context.tag[self.headers]

    def _get_site_url(self, secure):
        # secure base url factory -> get the corresponding config section
        if secure is True:
            _conf_section = Config().get_section("secure_server")
        else:
            _conf_section = Config().get_section("public_server")
        # get http connection type
        if Config().get("secure_server>use_ssl", type_=bool) is True:
            _http = "https"
        else:
            _http = "http"
        # little check
        if _conf_section is None:
            return None
        else:
            # get host and port
            _host = _conf_section["host"]
            _port = _conf_section["port"]
            # build url
            return "%s://%s:%s/%s%s"  \
                    % (_http, _host, _port, self.language_code, self.page_url)

    def _get_rpc_url(self, context):
        # get request object from context
        _request = inevow.IRequest(context)
        _r_type, _r_addr, _r_port = _request.getHost()
        # compare port to return good rpc url
        if _r_port == Config().get("secure_server>port"):
            return self._get_site_url(True)
        else:
            return self._get_site_url(False)

    def _add_widget(self, widget):
        # prepare
        widget.set_page(self)
        # do add
        self.widgets.append(widget)
        # update callbacks
        widget._update_callbacks()

    def render_widgets(self, context, data):
        """Return the doc factory content of the current widgets of the page.
        """
        # list of doc factories to render
        _factory_list = list()
        # site base urls
        _sec_url = self._get_site_url(True)
        _pub_url = self._get_site_url(False)
        # get rpc url
        _rpc_url = self._get_rpc_url(context)
        # add urls as hidden for ajax calls
        if _sec_url is None:
            pass
        else:
            # set secure url in page
            _sec_t = T.input(type_="hidden", id_="SecuredUrl", value=_sec_url)
            _factory_list.append(_sec_t)
        # public url base
        if _pub_url is None:
            pass
        else:
            # set secure url in page
            _pub_t = T.input(type_="hidden", id_="UnsecuredUrl", value=_pub_url)
            _factory_list.append(_pub_t)
        # rpc url base
        if _rpc_url is None:
            pass
        else:
            # set rpc url in page
            _rpc_t = T.input(type_="hidden", id_="RpcUrl", value=_rpc_url)
            _factory_list.append(_rpc_t)
        # keep data
        _data = inevow.IRequest(context).args
        # prepare and render
        for _w in self.widgets:
            # first realize
            _w.realize(_data)
            # init/update ajax callbacks
            _w._update_callbacks()
            # render
            _factory = _w.get_docFactory()
            # update factory list
            _factory_list.append(_factory)
        # return render
        return context.tag[_factory_list]

