#!/usr/bin/env python
# -*- coding: utf-8 -*-

# twisted import
from twisted.python import log

# nevow import
from nevow.i18n import _

# recaptcha import
from recaptcha.client import captcha

# bewype ui import
from bewype.core import ui
# bewype component import
from bewype.core.component.component import Component
# bewype config import
from bewype.core.config import Config
# bewype meassge import
from bewype.core.message import RpcMessage, RpcResult

class ReCaptcha(Component):

    def add_static_resources(self, file_=None):
        ui.Widget.add_static_resources(self, file_=__file__)

    def get_actions(self):
        return ["check"]

    def get_value_key(self):
        return "re_captcha"

    def get_field(self, item):
        # get server url from parent config
        _api_url = Config().get_value("recaptcha", "api_url")
        _auth_url = "%s/challenge" % _api_url
        # get public key from config
        _public_key = Config().get_value("recaptcha", "public_key")
        #
        return ui.ReCaptcha(class_=self.class_, auth_url=_auth_url,
                public_key=_public_key)

    def do_check(self, data, public=False):
        """ReCaptcha Errors:

        * *unknown* Unknown error.
        * *invalid-site-public-key* We weren't able to verify the public key.
        * *invalid-site-private-key* We weren't able to verify the private key.
        * *invalid-request-cookie* The challenge parameter of the verify script
          was incorrect.
        * *incorrect-captcha-sol* The CAPTCHA solution was incorrect.
        * *verify-params-incorrect* The parameters to verify were incorrect make
          sure you are passing all the required parameters.
        * *invalid-referrer* reCAPTCHA API keys are tied to a specific domain
          name for security reasons.
        * *recaptcha-not-reachable* reCAPTCHA never returns this error code. A
          plugin should manually return this code in the unlikely event that it
          is unable to contact the reCAPTCHA verify server.

        """
        # get avatar id
        _avatarId = self.page.get_avatar_id()
        # prepare result
        _r = RpcResult()
        # not connected
        if _avatarId is None \
        and public is False:
            _r.error(_("Not connected!"))
            return _r
        # is connected
        else:
            # parser message
            _m = RpcMessage(data)
            # parse fields values
            _f_challenge = _m.get_val("recaptcha_challenge")
            _f_response = _m.get_val("recaptcha_response")
            # little check
            if _f_response is None \
            or _f_response.strip() == "":
                # simple message for users
                _r.error(_("'Captcha' response field could not be empty!"))
                # reCaptcha result
                return _r
            else:
                # get private key from config
                _private_key = Config().get_value("recaptcha", "private_key")
                # call recaptcha server
                _response = captcha.submit(
                        _f_challenge,
                        _f_response,
                        _private_key,
                        None
                        )
                # simple message wrapping
                if _response.is_valid is False:
                    # simple message for users
                    _r.error(_("'Captcha' response does not match!"))
                else:
                    # OK
                    _r.success(_("You did it!"))
                # reCaptcha result
                return _r

    def do_reload(self, data, public=False):
        return RpcResult()

