#!/usr/bin/env python
# -*- coding: utf-8 -*-

# python import
from datetime import datetime
from hashlib import sha1
import os

# peak import
from peak.util.imports import importString

# sqlalchemy import
from sqlalchemy.orm import backref, mapper, relation, synonym

# twisted import
from twisted.python import log

# bewype utils import
from bewype.core import utils
# bewype db import
from bewype.core.db import SessionManager
# bewype config import
from bewype.core.config import Config as _Config

# DUMMY OBJ

class Dummy(object):

    def __init__(self, dummy_name=None, dummy_str=None, dummy_text=None,
            dummy_position=0, dummy_date=None, dummy_color=None,
            dummy_filename=None, parent_id=None, **kargs):
        self.dummy_name = dummy_name
        self.dummy_str = dummy_str
        self.dummy_text = dummy_text
        self.dummy_position = dummy_position
        # format and set date
        if dummy_date is None \
        or dummy_date.strip() == "":
            self.dummy_date = None
        else:
            _d_format = _Config().date_format()
            self.dummy_date = datetime.strptime(dummy_date, _d_format)
        # set color
        self.dummy_color = dummy_color
        # set filename
        self.dummy_filename = dummy_filename
        # ..
        self.parent_id = parent_id

    def id_in_letters(self):
        return utils.get_id_in_letters(self.dummy_id)

    def __repr__(self):
        return "<Dummy('%s', '%s', '%s', '%s', '%s')>" % (
                self.dummy_name,
                self.dummy_str,
                self.dummy_position,
                self.dummy_date,
                self.parent_id
                )

# PERMISSION OBJ

class User(object):
    """Simple User db item for the bewype core application.
    """

    def __init__(self, user_name=None, display_name=None, email_address=None,
            password=None, language_id=None, **kargs):
        """Creates a new db item for a new bewype user.
        """
        # set new db item values
        self.user_name = user_name
        self.display_name = display_name
        self.email_address = email_address
        self.password = password
        self.created = datetime.now()
        self.last_attempt = datetime.now()
        self.nb_of_attempt = 0
        self.language_id = language_id

    def id_in_letters(self):
        return utils.get_id_in_letters(self.user_id)

    def __repr__(self):
        return "<Config('%s', '%s', '%s', '%s')>" % (
                self.user_name,
                self.email_address,
                self.nb_of_attempt,
                self.language_id
                )

    def has_widget_permission(self, widget_element):
        """Check if user paremissions match with a given widget element
        permissions.

        :param widget_element: `WidgetElement` with permissions to compare
        :return: True if `User` and `WidgetElement` permissions match
        :rtype: `bool`
        """
        # check if works for all patterns
        if widget_element is None:
            return True
        # no perms set .. should be public
        elif len(widget_element.permissions) == 0:
            return True
        # check perms
        else:
            pass
        # check permission
        for _p in widget_element.permissions:
            if _p in self.permissions:
                continue
            else:
                return False
        # user have all required permissions
        return True

    @property
    def permissions(self):
        """Returns the user permissions according to its groups and permissions.
        """
        _perms = set()
        for _g in self.groups:
            _perms = _perms | set(_g.permissions)
        return _perms

    @classmethod
    def by_email_address(cls, email):
        """A class method that can be used to search users based on their email
        addresses since it is unique.
        """
        _session = SessionManager().session
        _query = _session.query(cls)
        _query = _query.filter(cls.email_address==email)
        return _query.first()

    @classmethod
    def by_user_name(cls, user_name):
        """A class method that permits to search users based on their
        user_name attribute.
        """
        _session = SessionManager().session
        _query = _session.query(cls)
        _query = _query.filter(cls.user_name==user_name)
        return _query.first()

    def _set_password(self, pass_):
        """Encrypts password on the fly using the encryption algo defined in
        the configuration.
        """
        algorithm = self.get_encryption_method()
        self._password = self.__encrypt_password(algorithm, pass_)

    def _get_password(self):
        """Returns encrypted password.
        """
        return self._password

    password = property(_get_password, _set_password)

    def __encrypt_password(self, algorithm, pass_):
        """Hash the given password with the specified algorithm. Valid values
        for algorithm are 'salted_sha1'. All other algorithm values will
        be essentially a no-op.
        """
        hashed_password = pass_

        if isinstance(pass_, unicode):
            password_8bit = pass_.encode("UTF-8")
        else:
            password_8bit = pass_

        if "salted_sha1" == algorithm:
            salt = sha1()
            salt.update(os.urandom(60))
            hash = sha1()
            hash.update(password_8bit + salt.hexdigest())
            hashed_password = salt.hexdigest() + hash.hexdigest()
        else:
            ValueError("Only salted_sha1 is implemented, not '%s'" % algorithm)

        # make sure the hased password is an UTF-8 object at the end of the
        # process because SQLAlchemy _wants_ a unicode object for
        # Unicode columns
        if not isinstance(hashed_password, unicode):
            hashed_password = hashed_password.decode("UTF-8")

        return hashed_password

    def get_encryption_method(self):
        """Returns the encryption method from the config.
        If None is set, or auth is disabled this will return None.
        """
        return "salted_sha1"

    def validate_password(self, pass_):
        """Check the password against existing credentials.
        this method _MUST_ return a boolean.

        :param pass_: the password that was provided by the user to try and
                      authenticate. This is the clear text version that we will
                      need to match against the (possibly) encrypted one in the
                      database.
        """
        algorithm = self.get_encryption_method()
        if "salted_sha1" == algorithm:
            hashed_pass = sha1()
            hashed_pass.update(pass_ + self.password[:40])
            return self.password[40:] == hashed_pass.hexdigest()

        else:
            ValueError('Only salted_sha1 is implemented, not "%s"' % algorithm)

class Group(object):
    """An ultra-simple group definition.
    """

    def __ini__(self):
        pass

class Permission(object):
    """A relationship that determines what each Group can do
    """

    def __ini__(self):
        pass

# LANGUAGE OBJ

class Language(object):
    """Simple Language db item for the bewype core application.
    """

    def __init__(self, language_name=None, display_name=None,
            language_code=None, **kargs):
        """Creates a new db item for a new bewype language.
        """
        # set new db item values
        self.language_name = language_name
        self.display_name = display_name
        self.language_code = language_code

    def id_in_letters(self):
        return utils.get_id_in_letters(self.language_id)

    def __repr__(self):
        return "<Language('%s', '%s', '%s')>" % (
                self.language_name,
                self.display_name,
                self.language_code
                )

class Country(object):
    def __ini__(self):
        pass

# LAYOUT OBJ

class Layout(object):

    def id_in_letters(self):
        return utils.get_id_in_letters(self.layout_id)

class LayoutType(object):
    def __ini__(self):
        pass

class Page(object):
    """Simple Page db item for the bewype core application.
    """

    def __init__(self, page_url=None, layout_id=None, **kargs):
        """Creates a new db item for a new bewype page.
        """
        # set values
        self.page_url = page_url
        # related ids
        self.layout_id = layout_id

    def id_in_letters(self):
        return utils.get_id_in_letters(self.page_id)

    def get_page_title(self, language_code):
        """Returns `PageTitle.display_name` for the current Page item and
        language_code passed as parameter.

        :param language_code: language_code of the expected title
        :return: page title for the current page and given language code or None
        :rtype: `str`
        """
        # read titles for the current page
        for _t in self.titles:
            # find one with the expected language code
            if _t.language is None \
            or _t.language.language_code == language_code:
                return _t.display_name
            # continue
            else:
                continue
        # no title found            
        return ""

    def __repr__(self):
        return "<Page('%s', '%s')>" % (
                self.page_url,
                self.layout_id,
                )

class PageTitle(object):
    """Simple PageTitle db item for the bewype core application.
    """

    def __init__(self, display_name=None, description=None, language_id=None,
            page_id=None, **kargs):
        """Creates a new db item for a new bewype page title.
        """
        # set values
        self.display_name = display_name
        self.description = description
        # related ids
        self.language_id = language_id
        self.page_id = page_id

    def id_in_letters(self):
        return utils.get_id_in_letters(self.page_title_id)

    def __repr__(self):
        return "<PageTitle('%s', '%s', '%s')>" % (
                self.display_name,
                self.language_id,
                self.page_id
                )

# WIDGET OBJ

class Widget(object):
    def __ini__(self):
        pass

    def id_in_letters(self):
        return utils.get_id_in_letters(self.widget_id)

    def get_component_cls(self):
        """Returns a non-initialized component `class` using the current widget
        name. If import fails, returns None

        :return: component `class`
        :rtype: `class`
        """
        # get widget name and class name for string import
        _pkg_name = self.widget_name.replace("_", ".")
        _cls_name = utils.get_class_name(self.widget_name)
        # get the widget class using Importing feature
        _import_str = "bewype.%s.%s" % (_pkg_name, _cls_name)
        # DEBUG
        log.msg("[model] get_component_cls - _import_str: %s" % _import_str)
        # DEBUG
        # try the import
        try:
            return importString(_import_str)
        except Exception, e:
            # ERROR
            log.err("[model.Widget] get_component_cls- e: %s" % e)
            return None

class WidgetElement(object):
    """WidgetElement db item, link between layout and widget of the the bewype
    core application.
    """

    def __init__(self, display_name=None, element_position=None, layout_id=None,
            widget_id=None):
        """Creates a new db item for a new bewype widget element.
        """
        # set values
        self.display_name = display_name
        self.element_position = element_position
        # related ids
        self.layout_id = layout_id
        self.widget_id = widget_id

    def id_in_letters(self):
        return utils.get_id_in_letters(self.widget_element_id)

    def __repr__(self):
        return "<WidgetElement('%s', '%s', '%s', '%s')>" % (
                self.display_name,
                self.element_position,
                self.layout_id,
                self.widget_id
                )

# CONFIG OBJ

class Config(object):
    """Simple Language db item for a widget element of the the bewype core
    application.
    """

    def __init__(self, config_name=None, config_section=None, config_key=None,
            value_type=None, value_text=None, language_id=None,
            widget_element_id=None, **kargs):
        """Creates a new db item for a new bewype config.
        """
        # set values
        self.config_name = config_name
        self.config_section = config_section
        self.config_key = config_key
        self.value_type = value_type
        self.value_text = value_text
        # related ids
        self.language_id = language_id
        self.widget_element_id = widget_element_id

    def id_in_letters(self):
        return utils.get_id_in_letters(self.config_id)

    def __repr__(self):
        return "<Config('%s', '%s', '%s', '%s', '%s', '%s')>" % (
                self.config_name,
                self.config_section,
                self.config_key,
                self.value_text,
                self.language_id,
                self.widget_element_id
                )

# RESOURCE OBJ

class ResourceType(object):
    def __ini__(self):
        pass

class Resource(object):
    def __ini__(self):
        pass

# MENU OBJ

class Menu(object):
    def __ini__(self):
        pass

class MenuElement(object):
    def __ini__(self):
        pass

def init_mapping():
    #
    _t = SessionManager().get_tables()
    # dummy mapping
    mapper(Dummy, _t["bewype_dummy"],
            properties={
                "parent": relation(
                    Dummy,
                    remote_side=[_t["bewype_dummy"].c.dummy_id],
                    backref="children"
                    )
                },
            order_by=_t["bewype_dummy"].c.dummy_name
            )
    # permission mapping
    mapper(Permission, _t["bewype_permission"])
    #
    mapper(Group, _t["bewype_group"],
            properties={
                "permissions": relation(
                    Permission,
                    secondary=_t["bewype_group_permission"],
                    backref="groups",
                    order_by=_t["bewype_permission"].c.display_name
                    )
                },
            order_by=_t["bewype_group"].c.display_name
            )
    #
    mapper(User, _t["bewype_user"],
            properties={
                "_password": _t["bewype_user"].c.password,
                "language": relation(
                    Language,
                    uselist=False,
                    order_by=_t["bewype_language"].c.display_name
                    ),
                "groups": relation(
                    Group,
                    secondary=_t["bewype_user_group"],
                    backref="users",
                    order_by=_t["bewype_group"].c.display_name
                    )
                },
            order_by=_t["bewype_user"].c.display_name
            )
    # layout mapping
    mapper(Page, _t["bewype_page"],
            properties={
                "layout": relation(
                    Layout,
                    uselist=False,
                    backref=backref(
                        "page",
                        uselist=False,
                        ),
                    order_by=_t["bewype_layout"].c.display_name
                    )
                },
            order_by=_t["bewype_page"].c.page_url
            )
    mapper(PageTitle, _t["bewype_page_title"],
            properties={
                "page": relation(
                    Page,
                    uselist=False,
                    backref=backref(
                        "titles",
                        cascade="all, delete",
                        ),
                    order_by=_t["bewype_page"].c.page_url
                    ),
                "language": relation(
                    Language,
                    uselist=False,
                    backref=backref(
                        "titles",
                        cascade="all, delete",
                        ),
                    order_by=_t["bewype_language"].c.display_name
                    )
                },
            order_by=_t["bewype_page_title"].c.display_name
            )
    #
    mapper(LayoutType, _t["bewype_layout_type"])
    mapper(Layout, _t["bewype_layout"],
            properties={
                "children": relation(
                    Layout,
                    backref=backref(
                        "parent",
                        remote_side=[_t["bewype_layout"].c.layout_id]
                        ),
                    order_by=_t["bewype_layout"].c.layout_position
                    ),
                "layout_type": relation(
                    LayoutType,
                    uselist=False,
                    order_by=_t["bewype_layout_type"].c.display_name
                    )
                },
            order_by=_t["bewype_layout"].c.display_name
            )
    # language mapping
    mapper(Country, _t["bewype_country"],
            order_by=_t["bewype_country"].c.display_name)
    mapper(Language, _t["bewype_language"],
            properties={
                "country": relation(
                    Country,
                    uselist=False,
                    cascade="all, delete",
                    backref=backref(
                        "language",
                        uselist=False,
                        ),
                    order_by=_t["bewype_country"].c.display_name
                    )
                },
            order_by=_t["bewype_language"].c.display_name
            )
    # widget mapping
    mapper(Widget, _t["bewype_widget"],
            order_by=_t["bewype_widget"].c.display_name)
    # widget_element mapping
    mapper(WidgetElement, _t["bewype_widget_element"],
            properties={
                "widget": relation(
                    Widget,
                    uselist=False,
                    backref=backref(
                        "widget_elements",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_widget"].c.display_name
                    ),
                "layout": relation(
                    Layout,
                    uselist=False,
                    backref="widget_elements",
                    order_by=_t["bewype_layout"].c.display_name
                    ),
                "permissions": relation(
                    Permission,
                    secondary=_t["bewype_widget_element_permission"],
                    order_by=_t["bewype_permission"].c.display_name
                    )
                },
            order_by=_t["bewype_widget_element"].c.display_name
            )
    # config mapping
    mapper(Config, _t["bewype_config"],
            properties={
                "widget_element": relation(
                    WidgetElement,
                    uselist=False,
                    backref=backref(
                        "config",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_widget_element"].c.display_name
                    )
                },
            order_by=[
                _t["bewype_config"].c.config_name,
                _t["bewype_config"].c.config_section,
                _t["bewype_config"].c.config_key,
                _t["bewype_config"].c.value_type
                ]
            )
    # resource mapping
    mapper(ResourceType, _t["bewype_resource_type"])
    mapper(Resource, _t["bewype_resource"],
            properties={
                "language": relation(
                    Language,
                    uselist=False,
                    backref=backref(
                        "resources",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_language"].c.display_name
                    ),
                "resource_type": relation(
                    ResourceType,
                    uselist=False,
                    backref=backref(
                        "resources",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_resource_type"].c.display_name
                    ),
                "widget_element": relation(
                    WidgetElement,
                    uselist=False,
                    backref=backref(
                        "resources",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_widget_element"].c.display_name
                    )
                },
            order_by=_t["bewype_resource"].c.display_name
            )
    # menu mapping
    mapper(Menu, _t["bewype_menu"],
            properties={
                "language": relation(
                    Language,
                    uselist=False,
                    backref=backref(
                        "menus",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_language"].c.display_name
                    ),
                "widget_element": relation(
                    WidgetElement,
                    uselist=False,
                    order_by=_t["bewype_widget_element"].c.display_name
                    )
                },
            order_by=_t["bewype_menu"].c.display_name
            )
    #
    mapper(MenuElement, _t["bewype_menu_element"],
            properties={
                "language": relation(
                    Language,
                    uselist=False,
                    backref=backref(
                        "menu_elements",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_language"].c.display_name
                    ),
                "menu": relation(
                    Menu,
                    uselist=False,
                    backref=backref(
                        "children",
                        uselist=True
                        ),
                    order_by=_t["bewype_menu"].c.display_name
                    ),
                "parent": relation(
                    MenuElement,
                    remote_side=[_t["bewype_menu_element"].c.menu_element_id],
                    backref="children",
                    order_by=_t["bewype_menu_element"].c.display_name
                    ),
                "resource": relation(
                    Resource,
                    uselist=False,
                    order_by=_t["bewype_resource"].c.file_name
                    ),
                "widget_element": relation(
                    WidgetElement,
                    uselist=False,
                    backref=backref(
                        "menu_elements",
                        cascade="all, delete"
                        ),
                    order_by=_t["bewype_widget_element"].c.display_name
                    )
                },
            order_by=_t["bewype_menu_element"].c.display_name
            )

