#!/usr/bin/env python
# -*- coding: utf-8 -*-

# twisted import
from twisted.python import log

# sqlalchemy import
from sqlalchemy.sql import and_, select as sql_select

# epistock db import
from bewype.core.db import model, SessionManager
# epistock utils import
from bewype.core.utils import get_key_name, get_class_name

class Query(object):
    """The Query class gives us simple ways to querying our database. It offers
    a basic layer between the orm stuff and the application needs.

    In SQLAlchemy should be replaced (not for me) all the querying stuff should
    be changed here and the application will work.
    """

    def __init__(self, orm_class_name):
        """Creates the Query object with the orm class name and prepares common
        orm object and sql table parameters for querying.

        :param orm_class_name: name of the orm object to manage, ex.: "User",
                               "Config", Etc.
        :param order_by_key: attribute name on the current orm object to use for
                             order_by our results
        """
        # get the table obj
        _table_name = "bewype_%s" % get_key_name(orm_class_name)
        # set default id key
        self.key_id = "%s_id" % get_key_name(orm_class_name)
        self.position_key = "%s_position" % get_key_name(orm_class_name)
        # orm values
        if hasattr(model, orm_class_name):
            # common values
            self.obj = getattr(model, orm_class_name)
            # session shortcut for the orm part
            self.session = SessionManager().session
            # set key col
            if hasattr(self.obj, self.key_id):
                self.key_col = getattr(self.obj, self.key_id)
            else:
                self.key_col = None
        else:
            self.obj = None
            self.key_col = None
        #
        self.table = SessionManager().get_table(_table_name)
        if self.table is None:
            self.engine = None
        else:
            # engine shortcut for the orm part
            self.engine = SessionManager().engine

    def create(self, **kargs):
        """Creates a new db item using the current orm class. Args stuff should
        match with the current object __init__ api.

        Here is how we create a dummy item::

            >>> _q = Query('Dummy')
            >>> _d = {
            ...     'dummy_name': u'test_create'
            ...     }
            >>> _new_item = _q.create(**_d)

        Just a little check::

            >>> len(_q.get_all()) == 1
            True

        :return: the new created db item
        :rtype: `object`
        :see: `model`
        """
        # do create
        _new_obj = self.obj(**kargs)
        try:
            # do add
            self.session.add(_new_obj)
            self.session.commit()
            # return the new created obj at the end
            return _new_obj
        except Exception, e:
            # do rollback
            self.session.rollback()
            # ERROR
            log.err("[query] create - e: %s" % e)
            # nothing to return
            return None

    def _order_by(self, query, order_by_name):
        """Updates passed query with order by statement if the current orm
        object contains the asked column.

        :param query: query to be updated
        :param order_by_name: name of the order_by column, ex.: "user_name",
        "config_name", Etc.
        :return: the ordered_by query
        :rtype: `sqlalchemy.orm.query.Query`
        """
        # no order by parameter
        if order_by_name is None:
            return query
        # check orm class has the expected column
        elif getattr(self.obj, order_by_name):
            _order_by_column = getattr(self.obj, order_by_name)
            return query.order_by(_order_by_column)
        # no column to order by
        else:
            return query

    def get_all(self, order_by_name=None):
        """Returns all the rows of the current orm object. If the order_by_name
        parameter is set, the result will be order by the corresponding column.

        Add 2 dummy items::

            >>> _q = Query('Dummy')
            >>> _q.delete_all()
            >>> _d1 = {
            ...     'dummy_name': u'test_get_all_1',
            ...     'dummy_str': u'test_get_all_str'
            ...     }
            >>> _i_1 = _q.create(**_d1)
            >>> _d2 = {
            ...     'dummy_name': u'test_get_all_2',
            ...     'dummy_str': u'test_get_all_str'
            ...     }
            >>> _i_2 = _q.create(**_d2)

        Query all::

            >>> _r = _q.get_all()

        Just a little check::

            >>> isinstance(_r, list)
            True

            >>> len(_r) == 2
            True

        :param order_by_name: name of the order_by column, ex.: "user_name",
                              "config_name", Etc.
        :return: the results represented by this Query as a list
        :rtype: `list`
        """
        # get query
        _query = self.session.query(self.obj)
        # set order by rule
        _query = self._order_by(_query, order_by_name)
        # return all
        return _query.all()

    def get_item(self, key_name, key_value):
        """Returns a single db item matching with the passed key_name and
        key_value.

        Returns None if the key_name or the key_value is empty.

        Returns None if the current obj do not have corresponding column for
        the key_name.

        Add a dummy item::

            >>> _q = Query('Dummy')
            >>> _d = {
            ...     'dummy_name': u'test_get_item'
            ...     }
            >>> _new_item = _q.create(**_d)

        Do query item::

            >>> _r = _q.get_item('dummy_name', u'test_get_item')

        Just a little check, the result item should match with the current query
        object::

            >>> _new_item == _r
            True

        :param key_name: column name in string for the single where clause,
                         ex.: "user_name", "config_name", Etc.
        :param key_value: value of the single where clause, ex.: "admin",
                          "config_one", Etc.
        :return: the result row of the query or None
        :rtype: `object`
        """
        # ensure valid key name
        if key_name is None \
        or not hasattr(self.obj, key_name) \
        or (isinstance(key_name, str) and key_name.strip() == ""):
            return None
        # ensure valid key value
        if key_value is None \
        or (isinstance(key_value, str) and key_value.strip() == ""):
            return None
        else:
            # get SA query
            _query = self.session.query(self.obj)
            # get key
            _key = getattr(self.obj, key_name)
            # filtering
            _query = _query.filter(_key==key_value)
            # return the first found or Exception
            return _query.first()

    def get_items(self, criteria_list=None, order_by_name=None):
        """Returns a list of the current obj rows according to the criteria_list
        passed as parameter. If the order_by_name is specified the result will
        be sorted according to the corresponding obj column.

        Add a dummy item::

            >>> _q = Query('Dummy')
            >>> _d1 = {
            ...     'dummy_name': u'test_get_items_1',
            ...     'dummy_str': u'test_get_items_str'
            ...     }
            >>> _i_1 = _q.create(**_d1)
            >>> _d2 = {
            ...     'dummy_name': u'test_get_items_2',
            ...     'dummy_str': u'test_get_items_str'
            ...     }
            >>> _i_2 = _q.create(**_d2)
            >>> _d3 = {
            ...     'dummy_name': u'test_get_items_3',
            ...     'dummy_str': u'_'
            ...     }
            >>> _i_3 = _q.create(**_d3)

        Do query item::

            >>> _crit_list = [_q.obj.dummy_str==u'test_get_items_str']
            >>> _r = _q.get_items(criteria_list=_crit_list)

        Just a little check, the result item should match with the current query
        object::

            >>> len(_r) == 2
            True

            >>> _r[0].dummy_str == u'test_get_items_str'
            True

        :param criteria_list: criteria for the query filtering
        :param order_by_name: name of the order_by column, ex.: "user_name",
                              "config_name", Etc.
        :return: the list of rows for the query
        :rtype: `list`
        """
        # get query
        _query = self.session.query(self.obj)
        # set filter
        if criteria_list is None:
            pass
        else:
            _query = _query.filter(and_(*criteria_list))
        # set order by rule
        _query = self._order_by(_query, order_by_name)
        # return all
        return _query.all()

    def commit(self):
        """Proceeds and ensures the commit stuff. Should be use as a commit
        shorcut.

        If commit fails, here we try a rollback to ensure application next
        queries.

        Returns True in case of success, Exception object otherwise.

        :return: the commit result: True or Exception
        :rtype: `object`
        """
        # do commit
        try:
            # do commit
            self.session.commit()
            # result Ok!
            return True
        # Oops!
        except Exception, e:
            # ERROR
            log.msg("[query] commit - error: %s" % e)
            # rollback
            self.session.rollback()
            # result Ko!
            return e

    def _remove(self, item):
        """Remove an existing item from the current session.

        For example with the following a dummy item::

            >>> _q = Query('Dummy')
            >>> _d = {
            ...     'dummy_name': u'test_remove'
            ...     }
            >>> _item = _q.create(**_d)

        Remove the db item::

            >>> _q._remove(_item)
            True

        Just a little check::

            >>> _q.get_item('dummy_name', u'test_remove') is None
            True

        :param item: db item to be removed
        :return: the commit result: True or Exception
        :rtype: `object`
        """
        # remove
        self.session.delete(item)
        # commit
        return self.commit()

    def remove_item(self, key_name, key_value):
        """Removes the corresponding item to the key_name and key_value passed
        as parameters. First it query the db item item with the get_item method.
        If no item found the method returns False, else the method removes it.

        For example with the following a dummy item::

            >>> _q = Query('Dummy')
            >>> _d = {
            ...     'dummy_name': u'test_remove_item'
            ...     }
            >>> _item = _q.create(**_d)

        Remove the db item::

            >>> _q.remove_item('dummy_name', u'test_remove_item')
            True

        Just a little check::

            >>> _q.get_item('dummy_name', u'test_remove_item') is None
            True

        :param key_name: column name in string for the single where clause,
                         ex.: "user_name", "config_name", Etc.
        :param key_value: value of the single where clause, ex.: "admin",
                          "config_one", Etc.
        :return: False if not found otherwise the remove result: True or Exception
        :rtype: `object`
        """
        # get item first
        _item = self.get_item(key_name, key_value)
        # no item found
        if _item is None:
            return False
        else:
            # do remove
            return self._remove(_item)

    def delete_all(self):
        """Removes all the rows of the current table.

        For example with the current Dummy table, it contains previous sample rows
        that we did not removed::

            >>> _q = Query('Dummy')
            >>> len(_q.get_all()) != 0
            True

        After delete_all, everything disappears::

            >>> _q.delete_all()
            >>> len(_q.get_all()) == 0
            True

        """
        self.delete(None)

    def delete(self, criteria_list):
        """Removes all the rows of the current table according to the
        criteria_list passed as parameter.

        For example we add two dummy items::

            >>> _q = Query('Dummy')
            >>> _d1 = {
            ...     'dummy_name': u'test_delete_1'
            ...     }
            >>> _item = _q.create(**_d1)
            >>> _d2 = {
            ...     'dummy_name': u'test_delete_2'
            ...     }
            >>> _item = _q.create(**_d2)

        First we prepare the criteria list to remove the expected item(s),
        then we proceed to the delete::

            >>> _crit_list = [_q.obj.dummy_name==u'test_delete_1']
            >>> _q.delete(_crit_list)

        Just a little check::

            >>> _q.get_item('dummy_name', u'test_delete_1') is None
            True

            >>> _q.get_item('dummy_name', u'test_delete_2') is not None
            True

        :param criteria_list: criteria for the query filtering
        :return: True or False
        :rtype: `bool`
        """
        # prepare statement
        _delete_statmt = self.table.delete()
        # filter on passde criteria
        if criteria_list is None:
            pass
        else:
            _delete_statmt = _delete_statmt.where(and_(*criteria_list))
        # get connection
        _conn = self.engine.connect()
        # excute update
        _conn.execute(_delete_statmt)
        # close the connection
        _conn.close()

    def exist(self, criteria_list):
        """Returns True if mathching items with the criteria_list are present in
        current table.

        For example with a single dummy item::

            >>> _q = Query('Dummy')
            >>> _d = {
            ...     'dummy_name': u'test_exist'
            ...     }
            >>> _item = _q.create(**_d)

        First we prepare the criteria list to remove the expected item(s),
        then the exist check looks like this::

            >>> _crit_list = [_q.obj.dummy_name==u'test_exist']
            >>> _q.exist(_crit_list)
            True

        :param criteria_list: criteria for the query filtering
        :return: True or False
        :rtype: `bool`
        """
        # prepare statement
        _select_statmt = self.table.select()
        # set criteria list
        if criteria_list is None:
            pass
        else:
            _select_statmt = _select_statmt.where(and_(*criteria_list))
        # get connection
        _conn = self.engine.connect()
        # excute select
        _select_result = _conn.execute(_select_statmt)
        # get first row (if exist)
        _first_row = _select_result.fetchone()
        # close the connection
        _conn.close()
        # if not None item exist
        return _first_row is not None

    def insert(self, value_dict):
        """Do insert using SA sql like method. Create new item in database for
        the current table and the dict parameter.

        For example, here how we can insert a dummy item using the sql way::

            >>> _q = Query('Dummy')
            >>> _d = {
            ...     'dummy_name': u'test_insert'
            ...     }
            >>> _q.insert(_d)
            True

        Just a little check::

            >>> _q.get_item('dummy_name', u'test_insert') is not None
            True

        :value_dict: dict of the attributes and values to be inserted in a new
                     row of the current table
        :see: `http://www.sqlalchemy.org/docs/06/sqlexpression.html`
        """
        # prepare statement
        _insert_statmt = self.table.insert()
        _insert_statmt = _insert_statmt.values(value_dict)
        # get connection
        _conn = self.engine.connect()
        try:
            # excute update
            _conn.execute(_insert_statmt)
        except Exception, e:
            # ERROR
            log.err("[query] insert - e: %s" % e)
            return False
        # close the connection
        _conn.close()
        return True

    def update(self, key_value, value_dict):
        # prepare statement
        _update_statmt = self.table.update()
        # set where clause using current key_col
        _update_statmt = _update_statmt.where(self.key_col==key_value)
        # set values
        _update_statmt = _update_statmt.values(value_dict)
        # get connection
        _conn = self.engine.connect()
        try:
            # excute update
            _conn.execute(_update_statmt)
        except Exception, e:
            # ERROR
            log.err("[query] update - e: %s" % e)
            return False
        # close the connection
        _conn.close()
        return True

    def select(self, criteria_list=None, column_list=None, order_by_list=None):
        """Do select using SA sql like method. If all the parameters are set to
        None, all the rows of the current table will be returned. In that case
        the row order depends on the model definition.

        The column_list specifies the expected columns in returns. With
        order_by_list parameter, precising the result expected order column by
        column, this method can be used for direct data rendering task,
        ex.: csv, table, Etc.

        For example with 2 dummy rows::

            >>> _q = Query('Dummy')
            >>> _d1 = {
            ...     'dummy_name': u'test_select_1',
            ...     'dummy_str': u'select_test_str',
            ...     }
            >>> _item = _q.create(**_d1)
            >>> _d2 = {
            ...     'dummy_name': u'test_select_2',
            ...     'dummy_str': u'select_test_str'
            ...     }
            >>> _item = _q.create(**_d2)
            >>> _d3 = {
            ...     'dummy_name': u'test_select_3',
            ...     'dummy_str': u'_'
            ...     }
            >>> _item = _q.create(**_d3)

        First we prepare the criteria list to select the expected item(s),
        then we select the expected rows::

            >>> _crit_list = [_q.obj.dummy_str==u'select_test_str']
            >>> len(_q.select(_crit_list)) == 2
            True

        :param criteria_list: criteria for the query filtering
        :param column_list: expected columns filtering
        :param order_by_list: columns order_by setting
        :return: list of SA sql select resulting rows
        :rtype: `list`
        :see: `http://www.sqlalchemy.org/docs/06/sqlexpression.html`
        """
        # prepare statement
        if column_list is None:
            _select_statmt = self.table.select()
        else:
            _select_statmt = sql_select(column_list)
        # set criteria list
        if criteria_list is None:
            pass
        else:
            _select_statmt = _select_statmt.where(and_(*criteria_list))
        # set order list
        if order_by_list is None:
            pass
        else:
            _select_statmt = _select_statmt.order_by(*order_by_list)
        # get connection
        _conn = self.engine.connect()
        # excute select
        _select_result = _conn.execute(_select_statmt)
        # return result
        _result = _select_result.fetchall()
        # close the connection
        _conn.close()
        #
        return _result

    def _get_related_column(self, key_name):
        # get column
        _c = self.table.columns[key_name]
        # get foreign column
        return _c.foreign_keys[0].column

    def _get_related_class_name(self, key_name):
        # get foreign column
        _f_c = self._get_related_column(key_name)
        # get foreign table name
        _t_name = _f_c.table.name
        # get foreign column
        return get_class_name(_t_name.replace("bewype_", ""))

    def get_related_items(self, key_name):
        # get class name first
        _cls_name = self._get_related_class_name(key_name)
        # create new query and retrieve children
        return Query(_cls_name).get_all()

    def get_related_item(self, key_name, key_value):
        # get class name first
        _cls_name = self._get_related_class_name(key_name)
        # get foreign column
        _f_c = self._get_related_column(key_name)
        # create new query and retrieve children
        return Query(_cls_name).get_item(_f_c, key_value)

