#!/usr/bin/env python
# -*- coding: utf-8 -*-

# python import
import os, shutil

# twisted import
from twisted.python import log

# nevow import
from nevow import tags as T

# bewype resource import
from bewype.core.resource import Path
# bewype resource import
from bewype.core.template import Template

class StaticUpdater(object):
    """This class aim to update static resources in the server file system and
    the page header tags. It can add a single resource or a full tree directory.

    If a resource is not present in the server static folder it will be copied
    from the current resource path or nearby python file static folder.

    At the opposite if the resource already exist, only the tag header is
    updated.
    """

    def __init__(self, page, py_file=None, widget_element=None):
        """Creates a new update with the common required values for an update.

        :param page: `rend.Page` object to add the tag headers
        :param py_file: __file__ object
        :see: `nevow.rend.Page`
        """
        # set page to be updated
        self.page = page
        # set current py file
        self.py_file = py_file
        # set current widget element
        self.widget_element = widget_element

        # init local resource and tmpl path variables
        if self.py_file is None:
            # no py file no tmpl path
            self.loc_tmpl_path = None
        else:
            # set local template path based on py file
            self.loc_tmpl_path = \
                    Path().get_tmpl_path_from_py_file(self.py_file)

    def get_widget_element(self, widget_element):
        """Returns current widget element.
        """
        return self.widget_element

    def set_widget_element(self, widget_element):
        """Updates current widget element.
        """
        self.widget_element = widget_element

    def get_page_headers_path_list(self):
        """Return a list of existing paths in the current page headers list.

        :return: list of the static paths in the page headers
        :rtype: `list`
        """
        # init result path list
        _path_list = list()
        # check header one by one
        for _h in self.page.headers:
            # get path in js like tag header
            if _h.attributes.has_key("src"):
                _path_list.append(_h.attributes["src"])
            # get path in css like tag header
            elif _h.attributes.has_key("href"):
                _path_list.append(_h.attributes["href"])
            # tag should not link to a resource
            else:
                continue
        # return path list at the end
        return _path_list

    def add_page_header(self, resource_path):
        """Add a new page header for a given resource path if not already in the
        current page header tags.

        :param resource_path: path to add to the current page headers list
        :see: `nevow.rend.Page`
        """
        # get resource type
        _resource_type = Path(path=resource_path).get_type()
        # header already set
        if self.page is None \
        or resource_path in self.get_page_headers_path_list():
            return
        # set tag and update header var
        elif _resource_type == "script":
            self.page._add_header(
                    T.script(type="text/javascript", src=resource_path))
        elif _resource_type == "css":
            self.page._add_header(
                    T.link(rel="stylesheet", type="text/css", href=resource_path))
        # nothing to add            
        else:
            pass

    def append_static_resource(self, resource_path, use_type=True):
        """Add a single resource to the page header and to the server static
        directory (if not already there).

        Such a call::

            append_static_resource("test/test.js")

        Should obtain::

            resource/script/test/test.js > static/script/test/test.js

        :param resource_path: resource relative path to be added
        :param use_type: do not use file type in path if set to False:
                         resource/test/test.js > static/test/test.js
        """
        # create path helper
        _p = Path(path=resource_path, language_code=self.page.language_code,
                py_file=self.py_file)
        # get paths
        _src_path = _p.get_resource_path(use_type=use_type)
        # set widget element
        _p.set_widget_element(self.widget_element)
        # dst paths
        _dst_path = _p.get_static_path(use_type=use_type, use_py_file=False)
        _web_path = _p.get_web_path(use_type=use_type)
        #
        # little check - here we only manage file
        if _src_path is None \
        or os.path.isdir(_src_path) \
        or not os.path.exists(_src_path):
            # quit
            return False
        # check if local file is already in the server static path
        elif os.path.exists(_dst_path):
            pass
        # copy the local resource to the server static path
        else:
            # get server static dest path
            # _dst_path = _p.get_static_path()
            # get the server dir
            _dst_dir, _f = os.path.split(_dst_path)
            # check dir exist or create it
            if not os.path.exists(_dst_dir):
                os.makedirs(_dst_dir)
            else:
                pass
            # copy resource file
            shutil.copyfile(_src_path, _dst_path)
        # update page header at the end
        self.add_page_header(_web_path)
        return True

    def __update_first_file_in_list(self, resource_list, dir_path, first_file):
        """
        """
        # nothing to update
        if first_file is None:
            pass
        # set file path to set first
        else:
            _first_file_path = os.path.join(dir_path, first_file)
        # udpate list
        if _first_file_path in resource_list:
            # ensure first
            resource_list.remove(_first_file_path)
            resource_list.insert(0, _first_file_path)
        else:
            pass

    def append_static_dir(self, dir_path="", first_file=None):
        """Add all the files of a directory to the page header and to the
        server static directory (if not already there).

        If the `first_file` parameter is set, the corresponding resource will
        be added in first.

        :param dir_path: directory relative path in a static source path to add
        :param first_file: first file to add to the page header if set
        """
        # get source dir
        _p = Path(path=dir_path, py_file=self.py_file)
        _resource_dir = _p.get_resource_dir()
        if _resource_dir is None:
            return
        else:
            _src_dir = os.path.join(_resource_dir, dir_path)
        # little check
        if os.path.exists(_src_dir):
            pass
        else:
            return
        # init list of resource files to be added
        _resource_list = list()
        # enter in the local static dir and add dirs or list resources
        for i_ in os.listdir(_src_dir):
            # get local path for path type check
            _src_path = os.path.join(_src_dir, i_)
            # get dest path for action
            _dst_path = os.path.join(dir_path, i_)
            # check is dir
            if os.path.isdir(_src_path):
                # recursive stuff
                self.append_static_dir(_dst_path)
            elif os.path.isfile(_src_path):
                # add single file to be managed
                _resource_list.append(_dst_path)
            else:
                continue
        # little check
        if first_file is None:
            pass
        else:
            # sorry for that
            self.__update_first_file_in_list(_resource_list, dir_path,
                    first_file.replace("__", ""))
            # import file first at the end
            self.__update_first_file_in_list(_resource_list, dir_path, first_file)
        # map list for single resources appending
        for _r in _resource_list:
            self.append_static_resource(_r, use_type=False)

    def append_tmpl_resource(self, key_dict, resource_path, use_type=True,
            widget_id=None):
        """Add a single templated resource to the page header and to the
        server static directory (if not already there).

        At first stream is generated with a local template file and the key
        dictionary passed as parameter. Then the stream is written to the
        static path under the good file type folder and corresponding path.

        Such a call::

            append_tmpl_resource(<key_dict>, "test/test.js-tmpl")

        Should obtain::

            tmpl/script/test/test.js-tmpl > static/script/test/test.js

        :param key_dict: keys to be replaced in the template
        :param use_type: do not use file type in path if set to False:
                         tmpl/test/test.js-tmpl > static/test/test.js
        :param resource_path: templated resource relative path to be added
        """
        # little check
        _s_name = resource_path.split("-tmpl")
        if len(_s_name) != 2 \
        or len(_s_name[1]) != 0:
            return False
        # set dest file name
        else:
            _file_path = _s_name[0]
        # create path helper
        _p = Path(path=_file_path, language_code=self.page.language_code,
                py_file=self.py_file)
        # get paths
        _src_path = "%s-tmpl" % _p.get_tmpl_path(use_type=use_type)
        # set widget element
        _p.set_widget_element(self.widget_element)
        # update path with class name specific param
        if widget_id is None:
            pass
        else:
            _ext = _p.get_ext()
            _p.set_path(_file_path.replace(".%s" % _ext,
                ".%s.%s" % (widget_id, _ext)))
        # dst paths
        _dst_path = _p.get_static_path(use_type=use_type, use_py_file=False)
        _web_path = _p.get_web_path(use_type=use_type)
        # render file template to dest file if doesn't exist yet
        if self.loc_tmpl_path is None \
        or _dst_path is None:
            return False
        if os.path.exists(_dst_path):
            pass
        else:
            # write the new templated resource
            if Template(_src_path).write(key_dict, _dst_path):
                pass
            # error during writing
            else:
                return False
        # update page header
        self.add_page_header(_web_path)
        # result flag
        return True

    def append_tmpl_dir(self, key_dict, dir_path="", widget_id=None):
        """Add all the files of a directory to the page header and to the
        server static directory (if not already there).

        For each single file calls the `append_tmpl_resource` for templating
        and copy to the server static path.

        For each directory do a recurvise tricks and continue on nex templates.

        Such a call::

            append_tmpl_resource(<key_dict>, dir_path="test")

        Should obtain::

            tmpl/test/test1.js-tmpl > static/test/test1.js
            tmpl/test/test2.js-tmpl > static/test/test2.js
            tmpl/test/_d/test3.js-tmpl > static/test/_d/test3.js
            Etc.

        :param key_dict: keys to be replaced in the template
        :param dir_path: directory relative path in a tmpl source path to add
        """
        # get source dir
        _p = Path(path=dir_path, py_file=self.py_file)
        _tmpl_dir = _p.get_tmpl_dir()
        # first check
        if _tmpl_dir is None:
            return False
        else:
            # get destination path
            _src_dir = os.path.join(_tmpl_dir, dir_path)
        # second check
        if os.path.exists(_src_dir):
            # enter in the tmpl dir and add dirs or list tmpl
            for i_ in os.listdir(_src_dir):
                # get local path for check
                _src_path = os.path.join(_src_dir, i_)
                # get dest path for action
                _dst_path = os.path.join(dir_path, i_)
                # check is dir
                if os.path.isdir(_src_path):
                    # recursive stuff
                    self.append_tmpl_dir(key_dict, _dst_path,
                            widget_id=widget_id)
                elif os.path.isfile(_src_path):
                    # add single file to be managed
                    self.append_tmpl_resource(key_dict, _dst_path,
                            use_type=False, widget_id=widget_id)
                else:
                    continue
        # src path does not exist                
        else:
            return False

