#!/usr/bin/env python
# -*- coding: utf-8 -*-

# python import
import unittest
#
from tempfile import TemporaryFile

# bewype db import
from bewype.core.db import Query
# bewype data import
from bewype.core.data import YamlConfigLoader, YamlDbLoader

# bewype tests import
from bewype.core.tests._db_manager import _DbManager
# init specific db manager singleton for tests
_DbManager()

class TestConfigLoader(unittest.TestCase):

    def setUp(self):
        # remove all dummy items for each test
        Query("Config").delete_all()

    def test_loading_good_file(self):
        # prepare yaml file obj like
        _f = TemporaryFile()
        _f.write("""
config_name_1:
    config_section_1:
        config_key_1_1:
            value_type: type_1_1
            value_text: text_1_1
    config_section_2:
        config_key_2_1:
            value_type: type_2_1
            value_text: text_2_1
""")
        _f.seek(0)
        # create config loader and load
        _conf_loader = YamlConfigLoader(0, 0, file_obj=_f)
        _conf_loader.load()

        # get query for check
        _q = Query("Config")

        # check all
        _r_list = _q.get_all()
        # must have 2
        assert len(_r_list) == 2, "Found: %s" % len(_r_list)

        # check config_name_1 config
        _crit_list = [
                _q.obj.config_name=="config_name_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 2
        assert len(_r_list) == 2, "Found: %s" % len(_r_list)

        # check config_section_1 section
        _crit_list = [
                _q.obj.config_name=="config_name_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 2
        assert len(_r_list) == 2, "Found: %s" % len(_r_list)

        # check config_key_1_1 section
        _crit_list = [
                _q.obj.config_key=="config_key_1_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 1
        assert len(_r_list) == 1, "Found: %s" % len(_r_list)

    def test_loading_bad_file(self):
        # prepare yaml file obj like
        _f = TemporaryFile()
        _f.write("""
config_name_1:
    config_section_1:
        config_key_1_1:
            value_type: type_1_1
            value_text: text_1_1
    config_section_2:
        config_key_2_1:
        config_key_2_2:
            value_type: type_2_2
            value_text: text_2_2
    config_section_3:
        config_key_3_1:
            value_text: text_3_1
        config_key_3_2:
            value_type: text_3_2
""")
        _f.seek(0)
        # create config loader and load
        _conf_loader = YamlConfigLoader(0, 0, file_obj=_f)
        _conf_loader.load()

        # get query for check
        _q = Query("Config")

        # check all
        _r_list = _q.get_all()
        # must have 2
        assert len(_r_list) == 2, "Found: %s" % len(_r_list)

        # check config_name_1 config
        _crit_list = [
                _q.obj.config_name=="config_name_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 2
        assert len(_r_list) == 2, "Found: %s" % len(_r_list)

        # check config_section_1 section
        _crit_list = [
                _q.obj.config_section=="config_section_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 0
        assert len(_r_list) == 1, "Found: %s" % len(_r_list)

        # check config_key_2_1 key
        _crit_list = [
                _q.obj.config_section=="config_section_2",
                _q.obj.config_key=="config_key_2_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 0
        assert len(_r_list) == 0, "Found: %s" % len(_r_list)

        # check config_key_2_2 key
        _crit_list = [
                _q.obj.config_section=="config_section_2",
                _q.obj.config_key=="config_key_2_2"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 1
        assert len(_r_list) == 1, "Found: %s" % len(_r_list)

        # check config_key_3_1 key
        _crit_list = [
                _q.obj.config_section=="config_section_3",
                _q.obj.config_key=="config_key_3_1"
                ]
        _r_list = _q.get_items(criteria_list=_crit_list)
        # must have 0
        assert len(_r_list) == 0, "Found: %s" % len(_r_list)

    def test_loading_existing_config(self):
        # prepare yaml file obj like
        _f = TemporaryFile()
        _f.write("""
config_name_1:
    config_section_1:
        config_key_1_1:
            value_type: type_1_1
            value_text: text_1_1
""")
        _f.seek(0)
        # create config loader and load
        _conf_loader = YamlConfigLoader(0, 0, file_obj=_f)
        # 1rst loading
        _conf_loader.load()
        # 2cd loading - Oops!
        _conf_loader.load()

        # get query for check
        _q = Query("Config")

        # check all
        _r_list = _q.get_all()

        # must have 1
        assert len(_r_list) == 1, "Found: %s" % len(_r_list)

class TestDbLoader(unittest.TestCase):

    def setUp(self):
        # remove all dummy items for each test
        Query("Dummy").delete_all()

    def test_loading_good_file(self):
        # prepare yaml file obj like
        _f = TemporaryFile()
        _f.write("""
Dummy:
    2:
        dummy_name: dummy_name_2
        dummy_str: dummy_str
    1:
        dummy_name: dummy_name_1
        dummy_str: dummy_str
""")
        _f.seek(0)
        # create config loader and load
        _db_loader = YamlDbLoader(file_obj=_f)
        _db_loader.load()

        # get query for check
        _q = Query("Dummy")

        # get all
        _r_list = _q.get_all()
        # check
        assert len(_r_list) == 2, "Found: %s" % len(_r_list)

        # get an existing item
        _r = _q.get_item("dummy_id", 1)
        # check item value
        assert _r.dummy_name == u"dummy_name_1", "Found: %s" % _r.dummy_name

        # get an existing item
        _r = _q.get_item("dummy_id", 2)
        # check item value
        assert _r.dummy_name == u"dummy_name_2", "Found: %s" % _r.dummy_name

    def test_loading_bad_file(self):
        # prepare yaml file obj like
        _f = TemporaryFile()
        _f.write("""
Dummy:
    2:
        dummy_name: dummy_name_2
        dummy_str: dummy_str
    1:
""")
        _f.seek(0)
        # create config loader and load
        _db_loader = YamlDbLoader(file_obj=_f)
        _db_loader.load()

        # get query for check
        _q = Query("Dummy")

        # get all
        _r_list = _q.get_all()
        # check
        assert len(_r_list) == 1, "Found: %s" % len(_r_list)

        # get an existing item
        _r = _q.get_item("dummy_id", 1)
        # check item value
        assert _r is None

        # get an existing item
        _r = _q.get_item("dummy_id", 2)
        # check item value
        assert _r.dummy_name == u"dummy_name_2", "Found: %s" % _r.dummy_name

    def test_loading_existing_item(self):
        # prepare yaml file obj like
        _f = TemporaryFile()
        _f.write("""
Dummy:
    1:
        dummy_name: dummy_name_1
        dummy_str: dummy_str
""")
        _f.seek(0)
        # create config loader and load
        _db_loader = YamlDbLoader(file_obj=_f)
        # 1rst laoding
        _db_loader.load()
        # 2cd laoding
        _db_loader.load()

        # get query for check
        _q = Query("Dummy")

        # get all
        _r_list = _q.get_all()
        # check
        assert len(_r_list) == 1, "Found: %s" % len(_r_list)

def suite():
    # create new test suite
    _suite = unittest.TestSuite()
    # add test ceses
    _suite.addTest(unittest.makeSuite(TestConfigLoader, "test_config_loader"))
    _suite.addTest(unittest.makeSuite(TestDbLoader, "test_db_loader"))
    # return test suite
    return _suite

