#!/usr/bin/env python
# -*- coding: utf-8 -*-

# python import
import gettext, os, shutil, unittest
#
from ConfigParser import SafeConfigParser

# babel import
from babel.messages import catalog

# bewype db import
from bewype.core.db import Query
# bewype component import
from bewype.core.component import Page, tags as T
# bewype config import
from bewype.core.config import Config
# bewype resource import
from bewype.core.resource import Path, Resource, I18nUpdater, StaticUpdater

def _removeI18TestFiles():
    # set .po file path
    _i18n_path = I18nUpdater("en").get_i18n_path()
    _dummy_path = os.path.join(_i18n_path, "dummy.po")
    _po_path = os.path.join(_i18n_path, "bewype.po")
    _mo_path = os.path.join(_i18n_path, "bewype.mo")
    # remove .po and .mo files
    Resource(_dummy_path).remove()
    Resource(_po_path).remove()
    Resource(_mo_path).remove()

def _removeStaticFiles():
    # get static dir
    _static_path = Config().static_path()
    # remove resource files
    Resource(os.path.join(_static_path, "css")).remove()
    Resource(os.path.join(_static_path, "script")).remove()
    Resource(os.path.join(_static_path, "ElementOne")).remove()

class TestI18nUpdater(unittest.TestCase):

    def setUp(self):
        # create config object
        _config = SafeConfigParser()
        # set minimal value
        _config.add_section("path")
        _config.set("path", "resource", "bewype/core/tests/resource")
        _config.set("path", "static", "bewype/core/tests/static")
        # init config singleton
        Config(config=_config, force=True)
        # clear python i18n cache
        gettext._translations = {}
        # clean
        _removeI18TestFiles()

    def tearDown(self):
        # clean
        _removeI18TestFiles()

    def test_get_i18n_dir(self):
        # create a new updater with language code 'en' and get i18n path
        _i18n_path = I18nUpdater("en").get_i18n_path()

        # get static dir
        _res_path = Config().resource_path()
        # build i18n path for check
        _check_path = os.path.join(_res_path, "i18n", "en", "LC_MESSAGES")

        # check path
        assert _i18n_path == _check_path, "Found: %s" % _i18n_path

    def test_write_po_file(self):
        # create updater object
        _updater = I18nUpdater("en")
        # get i18n path
        _i18n_path = _updater.get_i18n_path()
        # set .po file path
        _po_path = os.path.join(_i18n_path, "bewype.po")

        # init catalog to write
        _catalog = catalog.Catalog(locale="en")
        # udpate catalog
        _catalog.add("dummy_1", string="Dummy One")
        _catalog.add("dummy_2", string="Dummy Two")

        # create a new updater with language code 'en' and write .po file
        _updater.write_po_file(_catalog)

        # check .po path exists
        assert os.path.exists(_po_path)

        # read .po file
        _po_f = open(_po_path)
        _po_lines = _po_f.readlines()
        _po_f.close()

        # first message check
        assert "msgid \"dummy_1\"\n" in _po_lines
        assert "msgstr \"Dummy One\"\n" in _po_lines
        # second message check
        assert "msgid \"dummy_2\"\n" in _po_lines
        assert "msgstr \"Dummy Two\"\n" in _po_lines

    def test_merge_po_files(self):
        # create updater object
        _updater = I18nUpdater("en")
        # get i18n path
        _i18n_path = _updater.get_i18n_path()
        # set .po file path
        _po_path = os.path.join(_i18n_path, "bewype.po")

        # do merge .po files (test_one.po and test_two.po)
        _updater.merge_po_files()

        # check .po path exists
        assert os.path.exists(_po_path)

        # read .po file
        _po_f = open(_po_path)
        _po_lines = _po_f.readlines()
        _po_f.close()

        # first message check
        assert "msgid \"test_1\"\n" in _po_lines
        assert "msgstr \"Test One\"\n" in _po_lines
        # second message check
        assert "msgid \"test_2\"\n" in _po_lines
        assert "msgstr \"Test Two\"\n" in _po_lines

    def test_update_mo_file(self):
        # create updater object
        _updater = I18nUpdater("en")
        # get i18n path
        _i18n_path = _updater.get_i18n_path()
        # set .po file path
        _po_path = os.path.join(_i18n_path, "bewype.po")

        # do merge .po files (test_one.po and test_two.po)
        _updater.merge_po_files()

        # do merge .po files (test_one.po and test_two.po)
        _updater.update_mo_file(do_merge=False)

        # check .po path exists
        assert os.path.exists(_po_path)

        # clear python i18n cache
        gettext._translations = {}
        # prepare gettext domain
        _i18n_root_path = os.path.join(Config().resource_path(), "i18n")
        gettext.install("bewype", _i18n_root_path)
        # install translation
        _lang_en = gettext.translation("bewype", _i18n_root_path, languages=['en'])
        _lang_en.install()
        _ = _lang_en.gettext

        # simple gettext check -> has messages
        assert _("test_1") == "Test One", "Found: %s" % _("test_1")
        assert _("test_2") == "Test Two", "Found: %s" % _("test_1")
        # simple gettext check -> has no message yet
        assert _("dummy_1") == "dummy_1", "Found: %s" % _("dummy_1")
        assert _("dummy_2") == "dummy_2", "Found: %s" % _("dummy_2")

        # init catalog to write
        _catalog = catalog.Catalog(locale="en")
        # udpate catalog
        _catalog.add("dummy_1", string="Dummy One")
        _catalog.add("dummy_2", string="Dummy Two")

        # create a new updater with language code 'en' and write .po file
        _updater.write_po_file(_catalog)

        # set .po file path
        _dummy_po_path = os.path.join(_i18n_path, "dummy.po")
        # move dummy content to additional file to merge
        shutil.move(_po_path, _dummy_po_path)

        # do merge .po files (test_one.po and test_two.po)
        _updater.update_mo_file()

        # re-install translation
        _lang_en = gettext.translation("bewype", _i18n_root_path, languages=['en'])
        _lang_en.install()
        _ = _lang_en.gettext

        # simple gettext check -> has messages
        assert _("test_1") == "Test One", "Found: %s" % _("test_1")
        assert _("test_2") == "Test Two", "Found: %s" % _("test_1")
        # simple gettext check -> has no message yet
        assert _("dummy_1") == "Dummy One", "Found: %s" % _("dummy_1")
        assert _("dummy_2") == "Dummy Two", "Found: %s" % _("dummy_2")

class TestStaticUpdater(unittest.TestCase):

    def setUp(self):
        # create config object
        _config = SafeConfigParser()
        # set minimal value
        _config.add_section("application")
        _config.set("application", "default_language", "en")
        _config.add_section("path")
        _config.set("path", "resource", "bewype/core/tests/resource")
        _config.set("path", "static", "bewype/core/tests/static")
        # init config singleton
        Config(config=_config, force=True)
        # clean files
        _removeStaticFiles()
        # remove all dummy items for each test
        Query("WidgetElement").delete_all()
        Query("Page").delete_all()

    def tearDown(self):
        # clean files
        _removeStaticFiles()
        # remove all dummy items for each test
        Query("WidgetElement").delete_all()
        Query("Page").delete_all()

    def test_get_page_headers_path_list(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # create a static updater
        _updater = StaticUpdater(_page)

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should be empty
        assert len(_p_list) == 0, "Found: %s" % len(_p_list)

        # set more headers
        _page._add_header(T.title["This is a test!"])
        # css header
        _t = T.link(rel="stylesheet", type="text/css", href="/static/dummy.css")
        _page._add_header(_t)
        # js header
        _t = T.script(type="text/javascript", src="/static/dummy.js")
        _page._add_header(_t)

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_p_list) == 2, "Found: %s" % len(_p_listpath_list)
        assert "/static/dummy.css" in _p_list, "Found: %s" % _p_list
        assert "/static/dummy.js" in _p_list, "Found: %s" % _p_list

    def test_add_page_header(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # create a static updater
        _updater = StaticUpdater(_page)

        # add page headers using updater method
        _updater.add_page_header("/static/dummy.css")
        _updater.add_page_header("/static/dummy.js")
        # should not be managed
        _updater.add_page_header("/static/test.dummy")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_p_list) == 2, "Found: %s" % len(_p_list)
        assert "/static/dummy.css" in _p_list, "Found: %s" % _p_list
        assert "/static/dummy.js" in _p_list, "Found: %s" % _p_list

    def test_append_static_resource(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # create a static updater
        _updater = StaticUpdater(_page)

        # add page headers using updater method
        _updater.append_static_resource("dummy.css")
        _updater.append_static_resource("dummy.js")
        # should not be managed
        _updater.append_static_resource("test.dummy")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should be empty
        assert len(_p_list) == 0, "Found: %s" % len(_p_list)

        # add resource file and header
        _updater.append_static_resource("res_test.css")
        _updater.append_static_resource("res_test.js")
        # should not be managed
        _updater.append_static_resource("test.dummy")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_p_list) == 2, "Found: %s" % len(_p_list)
        assert "/static/css/res_test.css" in _p_list, "Found: %s" % _p_list
        assert "/static/script/res_test.js" in _p_list, "Found: %s" % _p_list
        assert os.path.exists("bewype/core/tests/static/css/res_test.css")
        assert os.path.exists("bewype/core/tests/static/script/res_test.js")

        # create a static updater with py file parameter
        # for nearby static resource
        _updater = StaticUpdater(_page, py_file=__file__)

        # add resource file and header
        _updater.append_static_resource("res_test.css")
        _updater.append_static_resource("res_test.js")
        # should not be managed
        _updater.append_static_resource("test.dummy")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_p_list) == 2, "Found: %s" % len(_p_list)
        assert "/static/css/res_test.css" in _p_list, "Found: %s" % _p_list
        assert "/static/script/res_test.js" in _p_list, "Found: %s" % _p_list
        assert os.path.exists("bewype/core/tests/static/css/res_test.css")
        assert os.path.exists("bewype/core/tests/static/script/res_test.js")

    def test_append_img_resource(self):
        # create widget element
        _q = Query("WidgetElement")
        # prepare insert dict
        _i_dict = {
                "widget_element_id": 1,
                "display_name": u"Name",
                "element_position": 0,
                "layout_id": 0,
                "widget_id": 0,
                }
        # insert first item
        _q.insert(_i_dict)
        # get widget element
        _we_item = _q.get_item("widget_element_id", 1)
        # create page element
        _q = Query("Page")
        # prepare insert dict
        _i_dict = {
                "page_id": 1,
                "page_url": u"/index",
                "layout_id": 0,
                }
        # insert first item
        _q.insert(_i_dict)

        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])
        # create a static updater
        _updater = StaticUpdater(_page, widget_element=_we_item)

        # add resource file and header
        _updater.append_static_resource("res_test.jpg")
        # should not be managed
        _updater.append_static_resource("test.dummy")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_p_list) == 0, "Found: %s" % _p_list
        assert os.path.exists(
                "bewype/core/tests/static/ElementOne/img/res_test.jpg")

    def test_append_default_static_dir(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])
        # create a static updater
        _updater = StaticUpdater(_page)

        # add dir resources and headers
        _updater.append_static_dir("css")

        # get path list
        _l = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_l) == 2, "Found: %s" % _l
        assert "/static/css/res_test.css" in _l, "Found: %s" % _l
        assert "/static/css/res_dir/res_test.css" in _l, "Found: %s" % _l
        assert os.path.exists(
                "bewype/core/tests/static/css/res_test.css")
        assert os.path.exists(
                "bewype/core/tests/static/css/res_dir/res_test.css")

        # add dir resources and headers
        _updater.append_static_dir("script")

        # get path list
        _l = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_l) == 4, "Found: %s" % _l
        assert "/static/script/res_test.js" in _l, "Found: %s" % _l
        assert "/static/script/res_dir/res_test.js" in _l, "Found: %s" % _l
        assert os.path.exists(
                "bewype/core/tests/static/script/res_test.js")
        assert os.path.exists(
                "bewype/core/tests/static/script/res_dir/res_test.js")

    def test_append_near_by_py_file_static_dir(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])
        # create a static updater with py file parameter
        # for nearby static dir
        _updater = StaticUpdater(_page, py_file=__file__)

        # add already existing dir headers
        _updater.append_static_dir("css")

        # get path list
        _l = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_l) == 2, "Found: %s" % _l
        assert "/static/css/res_test.css" in _l, "Found: %s" % _l
        assert "/static/css/res_dir/res_test.css" in _l, "Found: %s" % _l
        assert os.path.exists(
                "bewype/core/tests/static/css/res_test.css")
        assert os.path.exists(
                "bewype/core/tests/static/css/res_dir/res_test.css")

        # add already existing dir headers
        _updater.append_static_dir("script")

        # get path list
        _l = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_l) == 4, "Found: %s" % _l
        assert "/static/script/res_test.js" in _l, "Found: %s" % _l
        assert "/static/script/res_dir/res_test.js" in _l, "Found: %s" % _l
        assert os.path.exists(
                "bewype/core/tests/static/script/res_test.js")
        assert os.path.exists(
                "bewype/core/tests/static/script/res_dir/res_test.js")

    def _append_tmpl_resource_pass(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # prepare key_dict for templating stuff
        _key_dict = {"element_id": "TestOne"}

        # create a static updater with py file parameter
        # for nearby static resource
        _updater = StaticUpdater(_page, py_file=__file__)

        # add page headers using updater method
        _updater.append_tmpl_resource(_key_dict, "tmpl_test.css-tmpl")
        _updater.append_tmpl_resource(_key_dict, "tmpl_test.js-tmpl")
        _updater.append_tmpl_resource(_key_dict, "dummy.xx-tmpl")

        # prepare expected path
        _p_1 = "bewype/core/tests/static/css/tmpl_test.css"
        _p_2 = "bewype/core/tests/static/script/tmpl_test.js"

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_p_list) == 2, "Found: %s" % len(_p_list)
        assert "/static/css/tmpl_test.css" in _p_list, "Found: %s" % _p_list
        assert "/static/script/tmpl_test.js" in _p_list, "Found: %s" % _p_list
        assert os.path.exists(_p_1), "Not Found: %s" % _p_1
        assert os.path.exists(_p_2), "Not Found: %s" % _p_2

        # read content _p_1
        _f_1 = open(_p_1)
        _c_1 = _f_1.read()
        _f_1.close()
        # read content _p_2
        _f_2 = open(_p_2)
        _c_2 = _f_2.read()
        _f_2.close()

        # check contents
        assert "TestOne" in _c_1, "Found: %s" % _c_1
        assert "TestOne" in _c_2, "Found: %s" % _c_2

    def test_append_tmpl_resource(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # prepare key_dict for templating stuff
        _key_dict = {"element_id": "TestOne"}

        # create a static updater
        _updater = StaticUpdater(_page)

        # add page headers using updater method
        _updater.append_tmpl_resource(_key_dict, "tmpl_test.css-tmpl")
        _updater.append_tmpl_resource(_key_dict, "tmpl_test.js-tmpl")
        _updater.append_tmpl_resource(_key_dict, "dummy.xx-tmpl")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should empty
        assert len(_p_list) == 0, "Found: %s" % _p_list

        # 2 pass should be ok
        # -> tmpl files are only add on the first pass
        # 2cd pass should not change anything
        self._append_tmpl_resource_pass()
        self._append_tmpl_resource_pass()

    def _append_tmpl_dir_pass(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # prepare key_dict for templating stuff
        _key_dict = {"element_id": "TestOne"}

        # create a static updater with py file parameter
        # for nearby static resource
        _updater = StaticUpdater(_page, py_file=__file__)

        # add page headers using updater method
        _updater.append_tmpl_dir(_key_dict, dir_path="css")

        # prepare expected path
        _p_1 = "bewype/core/tests/static/css/tmpl_test.css"
        _p_2 = "bewype/core/tests/static/css/_d/_d_test.css"

        # get path list
        _l = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_l) == 2, "Found: %s" % len(_l)
        assert "/static/css/tmpl_test.css" in _l, "Found: %s" % _l
        assert "/static/css/_d/_d_test.css" in _l, "Found: %s" % _l
        assert os.path.exists(_p_1)
        assert os.path.exists(_p_2)

        # add page headers using updater method
        _updater.append_tmpl_dir(_key_dict, dir_path="script")

        # prepare expected path
        _p_3 = "bewype/core/tests/static/script/tmpl_test.js"
        _p_4 = "bewype/core/tests/static/script/_d/_d_test.js"

        # get path list
        _l = _updater.get_page_headers_path_list()
        # list should not be empty
        assert len(_l) == 4, "Found: %s" % len(_l)
        assert "/static/script/tmpl_test.js" in _l, "Found: %s" % _l
        assert "/static/script/_d/_d_test.js" in _l, "Found: %s" % _l
        assert os.path.exists(_p_3)
        assert os.path.exists(_p_4)

        # check contents
        for _p in [_p_3, _p_3, _p_3, _p_3]:
            # read content _p
            _f = open(_p)
            _c = _f.read()
            _f.close()
            # check content
            assert "TestOne" in _c, "Found: %s" % _c

    def test_append_tmpl_dir(self):
        # create a new page
        _page = Page(init_headers=False)
        # set dummy header
        _page._add_header(T.title["This is a test!"])

        # prepare key_dict for templating stuff
        _key_dict = {"element_id": "TestOne"}

        # create a static updater
        _updater = StaticUpdater(_page)

        # add page headers using updater method
        _updater.append_tmpl_dir(_key_dict)

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should empty
        assert len(_p_list) == 0, "Found: %s" % _p_list

        # 2 pass should be ok
        # -> tmpl files are only add on the first pass
        # 2cd pass should not change anything
        self._append_tmpl_dir_pass()
        self._append_tmpl_dir_pass()
        
        # add page headers using updater method
        _updater.append_tmpl_dir(_key_dict, dir_path="invalid")

        # get path list
        _p_list = _updater.get_page_headers_path_list()
        # list should not change
        assert len(_p_list) == 0, "Found: %s" % len(_p_list)

def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestI18nUpdater, "test_i18n_updater"))
    suite.addTest(unittest.makeSuite(TestStaticUpdater, "test_static_updater"))
    return suite

if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())

